@@ -6,7 +6,7 @@ import pandas as pd
 from .base_accessor import BaseAccessor
 from .features_utils import nanskew, nankurtosis, q25, q75, mode, modefreq, nextmode, nextmodefreq, lastmode, lastmodefreq, \
 	nanmin, nanmax, nanmean, nanstd, nanmedian, nanmaxmmin, nansum, nanskewabs, nankurtosisabs, q25abs, q75abs,  \
-	nanminabs, nanmaxabs, nanmeanabs, nanstdabs, nanmedianabs, nanmaxmminabs, nansumabs
+	nanminabs, nanmaxabs, nanmeanabs, nanstdabs, nanmedianabs, nanmaxmminabs, nansumabs, n_unique
 
 
 @pd.api.extensions.register_dataframe_accessor("kxy_features")
@@ -113,6 +113,8 @@ class FeaturesAccessor(BaseAccessor):
 			When specified, only rows with :code:`filter_target` greater than :code:`filter_target_gt` will be considered for feature generation.
 		filter_target_lt : str | None
 			When specified, only rows with :code:`filter_target` smaller than :code:`filter_target_gt` will be considered for feature generation.
+		include_filter_target : bool
+			Whether to use :code:`filter_target` for features generation.
 
 
 		Returns
@@ -137,44 +139,47 @@ class FeaturesAccessor(BaseAccessor):
 		# Aggregation of categorical variables
 		if cat_columns:
 			# Most frequent value and it's frequency of occurence
-			agg.update({'MODE(%s)' % col: (col, mode) for col in cat_columns})
-			agg.update({'MODEFREQ(%s)' % col: (col, modefreq) for col in cat_columns})
+			agg.update({'%s__%s.MODE()' % (entity, col): (col, mode) for col in cat_columns})
+			agg.update({'%s__%s.MODEFREQ()' % (entity, col): (col, modefreq) for col in cat_columns})
 
 			# Second most frequent value and it's frequency of occurence
-			agg.update({'NEXTMODE(%s)' % col: (col, nextmode) for col in cat_columns})
-			agg.update({'NEXTMODEFREQ(%s)' % col: (col, nextmodefreq) for col in cat_columns})
+			agg.update({'%s__%s.NEXTMODE()' % (entity, col): (col, nextmode) for col in cat_columns})
+			agg.update({'%s__%s.NEXTMODEFREQ()' % (entity, col): (col, nextmodefreq) for col in cat_columns})
 
 			# Least frequent value and it's frequency of occurence
-			agg.update({'LASTMODE(%s)' % col: (col, lastmode) for col in cat_columns})
-			agg.update({'LASTMODEFREQ(%s)' % col: (col, lastmodefreq) for col in cat_columns})
+			agg.update({'%s__%s.LASTMODE()' % (entity, col): (col, lastmode) for col in cat_columns})
+			agg.update({'%s__%s.LASTMODEFREQ()' % (entity, col): (col, lastmodefreq) for col in cat_columns})
+
+			# Number of unique values
+			agg.update({'%s__%s.NUM_UNIQUE()' % (entity, col): (col, n_unique) for col in cat_columns})
 
 
 		# Aggregation of ordinal variables
 		if ord_columns:
-			agg.update({'SUM(%s)' % col: (col, nansum) for col in ord_columns})
-			agg.update({'MEAN(%s)' % col: (col, nanmean) for col in ord_columns})
-			agg.update({'STD(%s)' % col: (col, nanstd) for col in ord_columns})
-			agg.update({'MEDIAN(%s)' % col: (col, nanmedian) for col in ord_columns})
-			agg.update({'SKEW(%s)' % col: (col, nanskew) for col in ord_columns})
-			agg.update({'KURT(%s)' % col: (col, nankurtosis) for col in ord_columns})
-			agg.update({'Q25(%s)' % col: (col, q25) for col in ord_columns})
-			agg.update({'Q75(%s)' % col: (col, q75) for col in ord_columns})
-			agg.update({'MIN(%s)' % col: (col, nanmin) for col in ord_columns})
-			agg.update({'MAX(%s)' % col: (col, nanmax) for col in ord_columns})
-			agg.update({'MAX(%s)-MIN(%s)' % (col, col): (col, nanmaxmmin) for col in ord_columns})
+			agg.update({'%s__%s.SUM()' % (entity, col): (col, nansum) for col in ord_columns})
+			agg.update({'%s__%s.MEAN()' % (entity, col): (col, nanmean) for col in ord_columns})
+			agg.update({'%s__%s.STD()' % (entity, col): (col, nanstd) for col in ord_columns})
+			agg.update({'%s__%s.MEDIAN()' % (entity, col): (col, nanmedian) for col in ord_columns})
+			agg.update({'%s__%s.SKEW()' % (entity, col): (col, nanskew) for col in ord_columns})
+			agg.update({'%s__%s.KURT()' % (entity, col): (col, nankurtosis) for col in ord_columns})
+			agg.update({'%s__%s.Q25()' % (entity, col): (col, q25) for col in ord_columns})
+			agg.update({'%s__%s.Q75()' % (entity, col): (col, q75) for col in ord_columns})
+			agg.update({'%s__%s.MIN()' % (entity, col): (col, nanmin) for col in ord_columns})
+			agg.update({'%s__%s.MAX()' % (entity, col): (col, nanmax) for col in ord_columns})
+			agg.update({'%s__%s.MAX-MIN()' % (entity, col): (col, nanmaxmmin) for col in ord_columns})
 
 			if mix_sgn_columns:
-				agg.update({'SUM(ABS(%s))' % col: (col, nansumabs) for col in mix_sgn_columns})
-				agg.update({'MEAN(ABS(%s))' % col: (col, nanmeanabs) for col in mix_sgn_columns})
-				agg.update({'STD(ABS(%s))' % col: (col, nanstdabs) for col in mix_sgn_columns})
-				agg.update({'MEDIAN(ABS(%s))' % col: (col, nanmedianabs) for col in mix_sgn_columns})
-				agg.update({'SKEW(ABS(%s))' % col: (col, nanskewabs) for col in mix_sgn_columns})
-				agg.update({'KURT(ABS(%s))' % col: (col, nankurtosisabs) for col in mix_sgn_columns})
-				agg.update({'Q25(ABS(%s))' % col: (col, q25abs) for col in mix_sgn_columns})
-				agg.update({'Q75(ABS(%s))' % col: (col, q75abs) for col in mix_sgn_columns})
-				agg.update({'MIN(ABS(%s))' % col: (col, nanminabs) for col in mix_sgn_columns})
-				agg.update({'MAX(ABS(%s))' % col: (col, nanmaxabs) for col in mix_sgn_columns})
-				agg.update({'MAX(ABS(%s))-MIN(ABS(%s))' % (col, col): (col, nanmaxmminabs) for col in mix_sgn_columns})
+				agg.update({'%s__%s.ABS().SUM()' % (entity, col): (col, nansumabs) for col in mix_sgn_columns})
+				agg.update({'%s__%s.ABS().MEAN()' % (entity, col): (col, nanmeanabs) for col in mix_sgn_columns})
+				agg.update({'%s__%s.ABS().STD()' % (entity, col): (col, nanstdabs) for col in mix_sgn_columns})
+				agg.update({'%s__%s.ABS().MEDIAN()' % (entity, col): (col, nanmedianabs) for col in mix_sgn_columns})
+				agg.update({'%s__%s.ABS().SKEW()' % (entity, col): (col, nanskewabs) for col in mix_sgn_columns})
+				agg.update({'%s__%s.ABS().KURT()' % (entity, col): (col, nankurtosisabs) for col in mix_sgn_columns})
+				agg.update({'%s__%s.ABS().Q25()' % (entity, col): (col, q25abs) for col in mix_sgn_columns})
+				agg.update({'%s__%s.ABS().Q75()' % (entity, col): (col, q75abs) for col in mix_sgn_columns})
+				agg.update({'%s__%s.ABS().MIN()' % (entity, col): (col, nanminabs) for col in mix_sgn_columns})
+				agg.update({'%s__%s.ABS().MAX()' % (entity, col): (col, nanmaxabs) for col in mix_sgn_columns})
+				agg.update({'%s__%s.ABS().MAX-MIN()' % (entity, col): (col, nanmaxmminabs) for col in mix_sgn_columns})
 		
 		# Number of rows per entity
 		agg.update({'COUNT(%s)' % entity_name: (columns[0], 'count')})
@@ -229,10 +234,10 @@ class FeaturesAccessor(BaseAccessor):
 		df = self._obj.copy()
 		if ord_columns:
 			for col in ord_columns:
-				df['ABS(%s - MEAN(%s))' % (col, col)] = np.abs(df[col]-means.loc[col])
-				df['ABS(%s - MEDIAN(%s))' % (col, col)] = np.abs(df[col]-quantiles.loc[0.5][col])
-				df['ABS(%s - Q25(%s))' % (col, col)] = np.abs(df[col]-quantiles.loc[0.25][col])
-				df['ABS(%s - Q75(%s))' % (col, col)] = np.abs(df[col]-quantiles.loc[0.75][col])
+				df['%s.ABS(* - MEAN(*))' % col] = np.abs(df[col]-means.loc[col])
+				df['%s.ABS(* - MEDIAN(*))' % col] = np.abs(df[col]-quantiles.loc[0.5][col])
+				df['%s.ABS(* - Q25(*))' % col] = np.abs(df[col]-quantiles.loc[0.25][col])
+				df['%s.ABS(* - Q75(*))' % col] = np.abs(df[col]-quantiles.loc[0.75][col])
 
 		if return_baselines:
 			return df, means, quantiles
@@ -240,7 +245,7 @@ class FeaturesAccessor(BaseAccessor):
 			return df
 
 
-	def temporal_features(self, max_lag=10, exclude=[], index=None):
+	def temporal_features(self, max_lag=10, exclude=[], index=None, groupby=None, sort_by=None):
 		"""
 		Extend the dataframe with some rolling statistics (e.g. rolling average, rolling min, rolling max, rolling max-rolling min, etc.) for all lags from 2 to the configured maximum lag.
 
@@ -253,6 +258,10 @@ class FeaturesAccessor(BaseAccessor):
 			The largest lag to consider.
 		index : str | None (default)
 			The column, if any, to set as index and sort before computing rolling statistics.
+		groupby : str | None
+			If provided, we will use this column to perform a groupby before temporal aggregation.
+		sort_by : str | list | None
+			Columns, if any, we need to sort the dataframe by, prior to rolling.
 
 
 		Returns
@@ -260,49 +269,89 @@ class FeaturesAccessor(BaseAccessor):
 		result : pandas.DataFrame
 			The original dataframe extended with computed temporal features.
 		"""
+		if index:
+			exclude += [index]
+
+		if groupby:
+			exclude += [groupby]
+
 		ord_columns = [col for col in self._obj.columns if not self.is_categorical(col) and col not in exclude]
 		df = self._obj.copy()
-		if index:
+		if index and (df.index.name != index):
 			df = df.set_index(index)
-		df = df.sort_index()
+
+		if sort_by:
+			df = df.sort_values(by=sort_by)
+
 		dfs = [df.copy()]
 
-		for lag in range(2, max_lag+2):
-			rol_grp = df.rolling(window=lag, min_periods=1)
-			lagged_mean_df = rol_grp.aggregate(nanmean).rename(columns={col: 'MEAN(%s, %d)' % (col, lag) for col in ord_columns})
-			dfs += [lagged_mean_df.copy()]
-			lagged_min_df = rol_grp.aggregate(nanmin).rename(columns={col: 'MIN(%s, %d)' % (col, lag) for col in ord_columns})
-			dfs += [lagged_min_df.copy()]
-			lagged_max_df = rol_grp.aggregate(nanmax).rename(columns={col: 'MAX(%s, %d)' % (col, lag) for col in ord_columns})
-			dfs += [lagged_max_df.copy()]
-			lagged_maxmmin_df = rol_grp.aggregate(nanmaxmmin).rename(columns={col: 'MAX(%s, %d)-MIN(%s, %d)' % (col, lag, col, lag) for col in ord_columns})
-			dfs += [lagged_maxmmin_df.copy()]
+		if groupby:
+			def apply_func(s):
+				''' '''
+				lag = max_lag+1
+				col_map = {}
+				col_map.update({'%s_nanmean' % col: '%s.GROUPBY(%s).LAST(%d).MEAN()' % (col, groupby, lag) for col in ord_columns})
+				col_map.update({'%s_nanmin' % col: '%s.GROUPBY(%s).LAST(%d).MIN()' % (col, groupby, lag) for col in ord_columns})
+				col_map.update({'%s_nanmax' % col: '%s.GROUPBY(%s).LAST(%d).MAX()' % (col, groupby, lag) for col in ord_columns})
+				col_map.update({'%s_nanmaxmmin' % col: '%s.GROUPBY(%s).LAST(%d).MAX-MIN()' % (col, groupby, lag) for col in ord_columns})
+				cols = [_ for _ in col_map.values()]
+				res = s.rolling(lag, min_periods=1).aggregate([nanmean, nanmin, nanmax, nanmaxmmin])
+				res.columns = res.columns.map('_'.join).to_series().map(col_map)
+				return res[cols]
 
-		df = pd.concat(dfs, axis=1)
+			feat_df = df.groupby(groupby, sort=False).apply(apply_func)
+			feat_df.reset_index(inplace=True)
+			if df.index.name:
+				feat_df.set_index(df.index.name, inplace=True)
+			else:
+				feat_df.set_index('level_1', inplace=True)
+				feat_df.index.name = df.index.name
+			feat_df = feat_df.drop(groupby, axis=1)
+			dfs += [feat_df]
+			df = pd.concat(dfs, axis=1)
+
+		else:
+			lag = max_lag+1
+			col_map = {}
+			col_map.update({'%s_nanmean' % col: '%s.LAST(%d).MEAN()' % (col, lag) for col in ord_columns})
+			col_map.update({'%s_nanmin' % col: '%s.LAST(%d).MIN()' % (col, lag) for col in ord_columns})
+			col_map.update({'%s_nanmax' % col: '%s.LAST(%d).MAX()' % (col, lag) for col in ord_columns})
+			col_map.update({'%s_nanmaxmmin' % col: '%s.LAST(%d).MAX-MIN()' % (col, lag) for col in ord_columns})
+			cols = [_ for _ in col_map.values()]
+			feat_df = df.rolling(lag, min_periods=1).aggregate([nanmean, nanmin, nanmax, nanmaxmmin])
+			feat_df.columns = feat_df.columns.map('_'.join).to_series().map(col_map)
+			dfs += [feat_df[cols]]
+			df = pd.concat(dfs, axis=1)
 
 		return df
 
 
 	def generate_features(self, entity=None, encoding_method='one_hot', index=None, max_lag=None, exclude=[], \
 			means=None, quantiles=None, return_baselines=False, entity_name='*', filter_target=None, \
-			filter_target_gt=None, filter_target_lt=None, include_filter_target=False):
+			filter_target_gt=None, filter_target_lt=None, include_filter_target=False, fill_na=False, \
+			temporal_groupby=None, temporal_sort_by=None):
 		"""
 		Generate a wide range of candidate features to search from.
 
-		We first compute entity features if needed. Then we extend the resulting dataframe with deviations of ordinal columns from row-wise aggregtes such as mean, median, 25th and 75th percentiles. Finally, we ordinally-encode the resulting dataframe and apply temporal transformations if required.
+		We first compute entity features if needed. 
+
+		Then we extend the resulting dataframe with deviations of ordinal columns from row-wise aggregtes such as mean, median, 25th and 75th percentiles. 
+
+		Finally, we ordinally-encode the resulting dataframe and apply temporal transformations if required.
 
 
 		Parameters
 		----------
 		entity : str
 			The column mapping rows to entities.
-
 		filter_target : str | None
 			When specified, this is a column based on which we need to restrict the dataframe before generating entity features.
 		filter_target_gt : str | None
 			When specified, only rows with :code:`filter_target` greater than :code:`filter_target_gt` will be considered for entity feature generation.
 		filter_target_lt : str | None
 			When specified, only rows with :code:`filter_target` smaller than :code:`filter_target_gt` will be considered for entity feature generation.
+		include_filter_target : bool
+			Whether to use :code:`filter_target` for features generation.
 		encoding_method : 'one_hot' (default) | 'binary'
 			The encoding method to use for categorical variables.
 		exclude : list
@@ -317,7 +366,10 @@ class FeaturesAccessor(BaseAccessor):
 			Which values, if any, to use as 25th, 50th, 75th percentiles for deviation features.
 		return_baselines : bool
 			Whether to return which baselines have been used for deviation features.
-
+		temporal_groupby : str | None
+			If provided, we will use this column to perform a groupby before temporal aggregation.
+		temporal_sort_by : str | None
+			If provided, we will use this column to sort the dataframe before rolling when computing temporal features.
 
 
 		Returns
@@ -333,6 +385,12 @@ class FeaturesAccessor(BaseAccessor):
 				include_filter_target=include_filter_target)
 			accessor = FeaturesAccessor(df)
 
+		if index:
+			exclude += [index]
+			
+		if temporal_groupby:
+			exclude += [temporal_groupby]
+
 		# Deviation features
 		res = accessor.deviation_features(exclude=exclude, means=means, quantiles=quantiles, return_baselines=return_baselines)
 		df = res[0] if return_baselines else res
@@ -344,7 +402,10 @@ class FeaturesAccessor(BaseAccessor):
 
 		if max_lag:
 			# Temporal/trend features
-			df = accessor.temporal_features(exclude=exclude, max_lag=max_lag, index=index)
+			df = accessor.temporal_features(exclude=exclude, max_lag=max_lag, index=index, groupby=temporal_groupby, sort_by=temporal_sort_by)
+
+		if fill_na:
+			df = df.fillna(df.median(skipna=True))
 
 		if return_baselines:
 			return df, res[1], res[2]
